#!/usr/bin/env python3
import argparse
import os.path
import sys
import time
import zlib
from collections import deque
from signal import signal, SIGINT

import requests
import urllib3

CHUNK_SIZES = {  # only values accepted by resumable upload
    '4MB': 4*1024*1024,  # 4194304
    '32MB': 32*1024*1024,  # 33554432
    '128MB': 128*1024*1024,  # 134217728
}
API_HOST = 'apis.uloz.to'
API_TOKEN = ';HG%7jW6@6/8vx">R;f('
RETRIES = 5
REQUEST_TIMEOUT = 30  # seconds
UPLOAD_TIMEOUT = 600  # seconds


# Parse script parameters
parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('files', help='Files to be uploaded.', nargs=argparse.REMAINDER)
parser.add_argument('-l', '--login', help='User account login at uloz.to website.', required=True)
parser.add_argument('--chunk-size', help='Upload file by chunks of this size.',
                    choices=list(CHUNK_SIZES.keys()), default='4MB')
parser.add_argument('--folder-slug', help='Slug of target folder. Uploaded files will be stored in the folder with'
                                          ' this slug. If none is given files will be stored in the'
                                          ' default root folder.')
parser.add_argument('--check-crc', help='Should the CRC32 checksum be verified.', action='store_true')
parser.add_argument('--verbose', help='Increase verbosity.', action='store_true')
parser.add_argument('-t', '--api-token', help='Unique application token. Can be obtained from uloz.to support.',
                    default=API_TOKEN)
parser.add_argument('--auth-token', help='Unique authentication token to be used instead of password. Can be obtained from uloz.to web.',
                    required=True)
parser.add_argument('--retry-slug', help='Private slug of an unfinished batch.')
parser.add_argument('--api-host', help='Ulozto/Pornfile API hostname.', default=API_HOST)
parser.add_argument('--ignore-ssl', help='Ignore SSL verification in requests.', action='store_true')

args = parser.parse_args()

if not args.files:
    print('No files to upload')
    sys.exit(0)


def print_request_verbose(response, print_request_body=True):
    if not args.verbose:
        return

    request = response.request
    request_headers_str = ', '.join(f'{key}:{val}' for key, val in request.headers.items())
    request_items = [f'{request.method} {request.url}', f'Headers: {request_headers_str}']
    if print_request_body:
        request_body = request.body.decode() if isinstance(request.body, bytes) else str(request.body)
        request_items.append(f'Body: {request_body}')
    request_str = 'Request: {}'.format('\n'.join(request_items))

    response_headers_str = ', '.join(f'{key}:{val}' for key, val in response.headers.items())
    response_str = f'Response: {response.status_code}\nHeaders: {response_headers_str}\nBody: {response.text}'
    print(f'{request_str}\n{response_str}')


def verbose_line_break():
    if args.verbose:
        print()


def rebuild_script_command_with_retry_args():
    if is_rerun:
        additional_args = []
    else:
        additional_args = ['--retry-slug', private_slug]
    input_files_start = len(sys.argv) - len(args.files)
    quoted_input_files = [f'"{file}"' for file in sys.argv[input_files_start:]]
    return ' '.join((sys.argv[0], *additional_args, *sys.argv[1:input_files_start], *quoted_input_files))


def sigint_handler(*_):
    if upload_signature and private_slug:
        print('Upload script interrupted')
        print(f'To continue, try running:\n{rebuild_script_command_with_retry_args()}')
    sys.exit(0)


def register_file_upload(check_crc):
    # generate checksum if enabled
    crc32_checksum = None
    if check_crc:
        with open(file_path, 'rb') as f:
            crc32_checksum = zlib.crc32(f.read())

    # register file for upload
    register_body = {'batch_file_id': batch_file_id,
                     'upload_signature': upload_signature,
                     'name': file_name,
                     'filesize': file_size,
                     'chunksize': chunk_size,
                     'crc32': crc32_checksum,
                     }
    register_response = requests.post(register_new_upload_url,
                                      verify=should_verify, timeout=REQUEST_TIMEOUT, json=register_body)
    print_request_verbose(register_response)
    register_response.raise_for_status()
    print(f'Registered file {batch_file_id} "{file_name}" for upload')
    verbose_line_break()


def upload_chunk(url_base, cid, count, bfid, data):
    chunk_upload_url = f'{url_base}/{cid}'
    for retry_number in range(1, RETRIES+1):
        print(f'Uploading chunk {cid}/{count} of file {bfid}. Try {retry_number}/{RETRIES}')

        try:
            chunk_upload_response = requests.post(chunk_upload_url,
                                                  verify=should_verify,
                                                  timeout=UPLOAD_TIMEOUT,
                                                  data=data,
                                                  headers={'Content-Type': 'application/octet-stream'})
        except Exception as exc:
            print(f'Chunk upload failed with: {exc}')
            continue

        try:
            chunk_upload_response.raise_for_status()
        except Exception as exc:
            print(f'Chunk upload failed with: {exc}')
        else:
            print(f'Chunk {chunk_id}/{chunk_count} of file {bfid} uploaded successfully')
            break
        finally:
            print_request_verbose(chunk_upload_response, print_request_body=False)
            verbose_line_break()
    else:
        print('Chunk upload out of retries')
        print(f'To continue, try running:\n{retry_command}')
        sys.exit(1)


def request_file_upload_status(slug, bfid):
    get_status_url = f'{upload_url_direct}/v1/chunked-file-upload/{slug}/{bfid}/status'
    response = session.get(get_status_url, verify=should_verify, timeout=REQUEST_TIMEOUT)
    print_request_verbose(response)
    verbose_line_break()
    return response


signal(SIGINT, sigint_handler)

should_verify = True
if args.ignore_ssl:
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    should_verify = False

is_rerun = args.retry_slug is not None
chunk_size = CHUNK_SIZES[args.chunk_size]
api_url = f'https://{args.api_host}'


# Set up session
session = requests.Session()
session.headers = {
    'X-Auth-Token': args.api_token,
    'Content-type': 'application/json',
    'Accept': 'application/json',
}


# Get user token
print(f"User {args.login} authentication")
login = {"login": args.login, "token": args.auth_token}
user_token_response = session.post(url=f'{api_url}/v5/auth/token', json=login, verify=should_verify)
if user_token_response.status_code == 400:
    print(f'Invalid authentication token or username!')
    sys.exit(1)
user_token_response.raise_for_status()
session.headers['X-User-Token'] = user_token_response.json()['token_id']
print('Authentication successful')
verbose_line_break()


# Renew upload + retrieve its signature / Get new upload slug+signature
if is_rerun:
    print('Refreshing upload batch + recovering upload signature')
    private_slug = args.retry_slug
    renew_body = {'user_login': args.login, 'private_slug': private_slug}
    renew_upload_response = session.post(f'{api_url}/v6/upload/link',
                                         verify=should_verify, timeout=REQUEST_TIMEOUT, json=renew_body)
    print_request_verbose(renew_upload_response)
    renew_upload_response.raise_for_status()
    upload_url_direct = renew_upload_response.json()['upload_resumable_base_url']
    upload_signature = renew_upload_response.json()['upload_signature']

    print('Updating upload validity')
    update_validity_body = {'upload_signature': upload_signature}
    update_validity_response = requests.post(f'{upload_url_direct}/v1/chunked-file-upload/{private_slug}',
                                             verify=should_verify, timeout=REQUEST_TIMEOUT, json=update_validity_body)
    print_request_verbose(update_validity_response)
    update_validity_response.raise_for_status()

    print('Upload renewed')
else:
    print('Fetching upload link')
    user = {'user_login': args.login}
    upload_link_response = session.post(f'{api_url}/v6/upload/link',
                                        verify=should_verify, timeout=REQUEST_TIMEOUT, json=user)
    print_request_verbose(upload_link_response)
    upload_link_response.raise_for_status()
    upload_url_direct = upload_link_response.json()['upload_resumable_base_url']
    upload_signature = upload_link_response.json()['upload_signature']
    private_slug = upload_link_response.json()['private_slug']
    print('Upload link fetched')
verbose_line_break()


# List the files to upload, unpacking one level of directories for convenience
files = list()
for path in args.files:
    if os.path.isfile(path):
        files.append(path)
    elif os.path.isdir(path):
        for name in os.listdir(path):
            in_dir_path = os.path.join(path, name)
            if os.path.isfile(in_dir_path):
                files.append(in_dir_path)
files = sorted(files)


# Upload files
retry_command = rebuild_script_command_with_retry_args()
try:
    register_new_upload_url = f'{upload_url_direct}/v1/chunked-file-upload'
    print(f'Uploading files through {register_new_upload_url}')
    batch_file_id_dict = dict()
    missing_chunks_for_file = dict()
    for batch_file_id, file_path in enumerate(files, start=1):
        batch_file_id_dict[file_path] = batch_file_id
        file_size = os.path.getsize(file_path)
        file_name = os.path.basename(file_path)

        # check if file: is complete / is missing chunks / needs to be registered
        registration_needed = True
        if is_rerun:
            status_response = request_file_upload_status(private_slug, batch_file_id)
            if status_response.status_code != 400:
                registration_needed = False
                status_response.raise_for_status()
                missing_chunks = status_response.json()['missing_chunks']
                missing_chunks_for_file[batch_file_id] = missing_chunks

        if registration_needed:
            register_file_upload(args.check_crc)

        # send file chunk by chunk
        print(f'Uploading file {batch_file_id} "{file_name}" using chunk size {args.chunk_size} ({chunk_size} B)')
        chunk_count = file_size // chunk_size
        if file_size % chunk_size != 0:
            chunk_count += 1
        chunk_upload_url_base = f'{upload_url_direct}/v1/chunked-file-upload/{private_slug}/{batch_file_id}'
        with open(file_path, 'rb') as f:
            chunk_id = 1
            chunk_data = f.read(chunk_size)
            while chunk_data:
                # skip already uploaded chunks
                if not registration_needed \
                        and chunk_id not in missing_chunks_for_file[batch_file_id]:
                    chunk_data = f.read(chunk_size)
                    chunk_id += 1
                    continue

                upload_chunk(chunk_upload_url_base, chunk_id, chunk_count, batch_file_id, chunk_data)

                chunk_data = f.read(chunk_size)
                chunk_id += 1
        time.sleep(1)

    # Wait for uploads to finish and gather data used to set file properties
    print('Waiting for uploads to finish')
    file_slugs = dict()
    upload_tokens = dict()
    waiting_queue = deque(files)
    while waiting_queue:
        file_path = waiting_queue.popleft()
        batch_file_id = batch_file_id_dict[file_path]

        # get file upload status
        status_response = request_file_upload_status(private_slug, batch_file_id)

        response_json = status_response.json()
        file_upload_status = response_json['status']
        print(f'File {batch_file_id}, Status: {file_upload_status}')
        if file_upload_status not in ('pending', 'processing'):
            if file_upload_status == 'finished':
                file_slug = response_json['irpc']['slug']
                file_slugs[file_slug] = file_path
                upload_tokens[file_slug] = f'{private_slug}:{batch_file_id}'
                print(f'File {batch_file_id} ({file_path}) uploaded successfully')
            elif file_upload_status == 'bad-crc32':
                print(f'File CRC32 checksum mismatch! bfid: {batch_file_id}, file: {file_path}')
            else:
                print(f'File {batch_file_id} ({file_path}) upload unsuccessful')
        else:
            waiting_queue.append(file_path)
        verbose_line_break()

        time.sleep(2)

    # If no files uploaded correctly, the rest is unnecessary
    if not file_slugs:
        print('No files uploaded successfully')
        sys.exit(1)

    # Set file properties
    print('Setting file properties')
    batch_properties = {'slugs': list(file_slugs.keys()),
                        'upload_tokens': upload_tokens,
                        }
    if args.folder_slug:
        batch_properties['folder_slug'] = args.folder_slug
    else:
        # get root folder slug
        session_response = session.get(f'{api_url}/v5/me', verify=should_verify, timeout=REQUEST_TIMEOUT)
        print_request_verbose(session_response)
        session_response.raise_for_status()
        batch_properties['folder_slug'] = session_response.json()['user']['root_folder_slug']

    properties_response = session.patch(url=f'{api_url}/v8/file-list/private', verify=should_verify, json=batch_properties)
    print_request_verbose(properties_response)
    properties_response.raise_for_status()
    print('File properties set')
    verbose_line_break()

    # Commit batch
    print('Committing files')
    commit = {'status': 'confirmed', 'owner_login': args.login}
    commit_batch_response = session.patch(f'{api_url}/v8/upload-batch/private/{private_slug}',
                                          verify=should_verify, timeout=REQUEST_TIMEOUT, json=commit)
    print_request_verbose(commit_batch_response)
    verbose_line_break()
    commit_batch_response.raise_for_status()

    # Generate after upload urls
    site = '.'.join(api_url.split('.')[1:])
    print('Files committed:')
    for file_slug, file_path in file_slugs.items():
        print(f'https://{site}/file/{file_slug}/{os.path.basename(file_path)}')

except Exception as exc:
    print(f'Upload interrupted by: {exc}')
    print(f'To continue, try running:\n{retry_command}')
    sys.exit(1)
